/*****************************************************************************
|*
|*  File:    serial.c
|*
|*  Version:    1.14
|*
|*  copyright 2001-2002 Altium BV
|*
 ****************************************************************************/

#include "target.h"
#include "break.h"
#include "serial.h"

static _xdat unsigned char se_buf[INPUT_BUFFER_SIZE+1];
static _xdat char se_buf_in;
static _xdat char se_buf_out;
static _xdat unsigned char flag_xoff_sent = 0;
static _xdat unsigned char flag_stopped = 0;
static _xdat unsigned char flag_send_xoff = 0;

#ifdef COPY_SERIAL_VECTOR
static _xdat unsigned char serial_vector[3] _at(SERIAL_VECTOR_ADDRESS);
#endif
static void clear_buffer(void);

/*****************************************************************************
|*
|*  FUNCTION:           se_int_transmit
|*
 */
_inline void
se_int_transmit(void)
{
    CLEAR_TI;
    if (flag_send_xoff)
    {
                                           /* we have to send XOFF */
        flag_send_xoff   = 0;
        WRITE_SERIAL_BUF = XOFF;
        flag_xoff_sent   = 1;    /* remember that we sent XOFF */
    }
    else
    {
        flag_TI = 1;
    }
}

/*****************************************************************************
|*
|*  FUNCTION:           se_int_receive
|*
|*  DESCRIPTION:
|*      receive interrupt: store character in circular input buffer
|*
 */
_inline void
se_int_receive( void )
{
    unsigned char x;
    char size;
    #pragma asm
        PUSH AR0
        PUSH AR1
        PUSH AR2
        PUSH AR3
        PUSH AR4
        PUSH AR5
        PUSH AR6
        PUSH AR7
    #pragma endasm
    READ_SERIAL_BUF(x);
    switch (x)
    {
    case INTR:
    case '!':
        flag_ctrlC = 1;              /* set flag for handler */
        flag_mon_intr = 1;        /* forced interrupt for handling ^C or single step */
        if (flag_TI)
        {
            FORCE_INTERRUPT;
        }
        break;
    case XOFF:
        flag_stopped = 1;
        break;
    case XON:
        flag_stopped = 0;
        break;
    default:
        /* store character in input buffer */
        se_buf[se_buf_in] = x;
        /* increment "in" pointer */
        if (++se_buf_in == INPUT_BUFFER_SIZE + 1)
        {
            se_buf_in = 0;        /* wrap around */
        }
        size = se_buf_in - se_buf_out;
        if (size < 0)
        {
            size += INPUT_BUFFER_SIZE+1;  /* wrap around */
        }
        if (!size) /* check for buffer overflow (se_buf_in == se_buf_out) */
        {
            /* buffer overflow! */
            se_buf_in--;
            if (se_buf_in < 0)
            {
                se_buf_in = INPUT_BUFFER_SIZE+1;  /* wrap around */
            }
        }
        else if (size == HIGH_WATER_MARK) /* size can be negative.... */
        {
            /* high water mark reached, stop sender */
            if (flag_TI)
            {
                flag_TI = 0;
                WRITE_SERIAL_BUF = XOFF;
                flag_xoff_sent   = 1;    /* remember that we sent XOFF */
            }
            else
            {
                flag_send_xoff = 1;    /* send XOFF as soon as transmitter is ready */
            }
        }
        break;
    }
#pragma asm
    POP AR7
        POP AR6
        POP AR5
        POP AR4
        POP AR3
        POP AR2
        POP AR1
        POP AR0
#pragma endasm
}

static void
clear_buffer_xon(void)
{
    WRITE_SERIAL_BUF = XON;
    flag_xoff_sent= 0;
}

static void
clear_buffer_wait(void)
{
    SERVICE_WATCHDOG;

    while (!flag_TI)
    {
        /* NOP */;
    }
    flag_TI = 0;
}

_frame(DPTR,PSW,ACC) _interrupt(SERIAL_INTERRUPT_NUMBER) void
se_interrupt(void)
{
    ALLOC_SCON;
    READ_SCON;

    if (READ_RI)
    {
        se_int_receive();
    }
    else if (flag_mon_intr)
    {
        flag_mon_intr = 0;
        RESET_INTERRUPT;    /* reset forced interrupt request */
        flag_TI = 1;
#pragma    asm
        EXTRN CODE(_?br_interrupt)
        LJMP _?br_interrupt;    /* perform a LJMP to save stackspace */
#pragma endasm
    }
    else if (READ_TI)
    {
        se_int_transmit();
    }
    return;
}

void
se_initialize_port(void)
{
#ifdef COPY_SERIAL_VECTOR
    unsigned int addr_serial_interrupt = (unsigned int)se_interrupt;

# ifndef HAS_AUXRAM
    serial_vector[0] = OPCODE_LJMP;
    serial_vector[1] = (unsigned char)((addr_serial_interrupt >> 8) & 0x00FF);
    serial_vector[2] = (unsigned char)(addr_serial_interrupt & 0x00FF);

# else
    /* HAS AUXRAM ,Use ta_write_auxcode */
    ta_write_auxcode(     SERIAL_VECTOR_ADDRESS   , OPCODE_LJMP );
    ta_write_auxcode(     SERIAL_VECTOR_ADDRESS+1 , (unsigned char)(((unsigned int)se_interrupt) >> 8));
    ta_write_auxcode(     SERIAL_VECTOR_ADDRESS+2 , (unsigned char)se_interrupt );

# endif

#endif

    flag_TI = 0;
    /* initialize serial port registers */
    ta_serial_init();
    clear_buffer();
}

static void
clear_buffer(void)
{
    se_buf_in  = 0;
    se_buf_out = 0;
    flag_stopped   = 0;    /* no XOFF received */
    flag_send_xoff = 0;    /* no pending XOFF request */
    if (flag_xoff_sent)
    {
        SERVICE_WATCHDOG;
        if ( flag_TI )
        {
            flag_TI = 0;
            clear_buffer_xon();
        }
        else
        {
            clear_buffer_wait();
        }
    }
}

void
se_putchar (unsigned char ch)
{
    do
    {
        SERVICE_WATCHDOG;
        while ( flag_stopped )
        {
            /* NOP */;
        }
    }
    while (!flag_TI);
    flag_TI = 0;
    WRITE_SERIAL_BUF = ch;
    if (ch == '\n')
    {
        se_putchar(RETURN);
    }
}

unsigned char
se_getchar(void)
{
    char x;
    do
    {
        SERVICE_WATCHDOG;
        x = se_buf_in - se_buf_out;    /* calculate current size of input buffer */
        if (x < 0)
        {
            x += INPUT_BUFFER_SIZE+1;  /* wrap around */
        }
        if ((x == LOW_WATER_MARK) && flag_xoff_sent)
        {
            /* wait until transmitter is available */
            while (!flag_TI)
            {
                ;
            }
            flag_TI = 0;
            WRITE_SERIAL_BUF = XON;    /* send XON even when 'flag_stopped' is set */
            flag_xoff_sent = 0;
        }
    } while (!x);                   /* wait until a character is available */
    x = se_buf[se_buf_out]; /* get character from serial buffer */
    se_buf_out++;              /* increment "out" pointer */
    if (se_buf_out == INPUT_BUFFER_SIZE+1)
    {
        se_buf_out = 0;        /* wrap around */
    }
    return x;
}

/*****************************************************************************
|*
|*  FUNCTION:           se_getline
|*
|*  DESCRIPTION:
|*
|*      Return the command line in the 'cmdline' buffer
|*
 */
void
se_getline(void)
{
    unsigned char x;
    unsigned char ptr = 0;
    do
    {
        x = se_getchar();    /* get one input character */
        switch (x)
        {
        case DEL:             /* treat DEL like BACKSPACE */
        case BACKSPACE:
            --P1_0;
            if (ptr > 0)
            {
                if (flag_echo_on)
                {
                    se_putchar(BACKSPACE);
                    se_putchar(' ');
                    se_putchar(BACKSPACE);
                }
                ptr--;
            }
            break;
        case ESC:
            while (ptr)
            {
                if (flag_echo_on)
                {
                    se_putchar(BACKSPACE);
                    se_putchar(' ');
                    se_putchar(BACKSPACE);
                }
                ptr--;
            }
            break;
        case RETURN:
            if (flag_echo_on)
            {
                se_putchar('\n');    /* echo RETURN */
            }
            cmdline[ptr] = '\0';
            break;
        default:
            if ((ptr < COMMAND_LINE_SIZE-1) /* don't write outside cmdline buffer */
                &&
                (x >= 0x20)  /* ignore control characters */
                )
            {
                if (flag_echo_on)
                {
                    se_putchar(x);    /* echo character */
                }
                cmdline[ptr++] = x;
            }
        }
    } while (x != RETURN);
    cmdline_ptr = cmdline;
}

/*****************************************************************************
|*
|*  FUNCTION:           se_print
|*
|*  DESCRIPTION:
|*      Output string + return
|*
 */
void
se_print(char *str)
{
    while (*str)
    {
        se_putchar(*str);
        str++;
    }
    se_putchar('\n');
}

